"""CLI commands for bin/cmempy-cli."""

import click

from tabulate import tabulate

from cmem.cmemc.cli import completion
from cmem.cmempy.queries import QUERY_CATALOG


@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only query identifier and no labels or other meta data. "
         "This is useful for piping the ids into other cmemc commands."
)
@click.command(name="list")
@click.pass_obj
def list_command(app, id_only):
    """List available queries from the catalog.

    Outputs a list of query URIs which can be used as reference for
    the query execute command.
    """
    if id_only:
        for _, sparql_query in QUERY_CATALOG.get_queries().items():
            app.echo_info(
                sparql_query.short_url
            )
    else:
        table = []
        for _, sparql_query in QUERY_CATALOG.get_queries().items():
            row = [
                sparql_query.short_url,
                sparql_query.query_type,
                ','.join(sparql_query.get_placeholder_keys()),
                sparql_query.label
            ]
            table.append(row)
        app.echo_info(
            tabulate(
                table,
                headers=["Query URI", "Type", "Placeholder", "Label"]
            )
        )


# pylint: disable-msg=too-many-locals,too-many-arguments
@click.command(name="execute")
@click.argument(
    "QUERIES",
    nargs=-1,
    required=True,
    autocompletion=completion.remote_queries_and_sparql_files
)
@click.option(
    "--accept",
    default="default",
    show_default=True,
    help="Accept header for the HTTP request(s). Setting this to 'default' "
         "means that cmemc uses an appropriate accept header for terminal "
         "output (text/csv for tables, text/turtle for graphs, * otherwise). "
         "Please refer to the Corporate Memory system manual for a list "
         "of accepted mime types."
)
@click.option(
    "--no-imports",
    is_flag=True,
    help="Graphs which include other graphs (using owl:imports) will be"
         "queried as merged overall-graph. This flag disables this"
         "default behaviour. The flag has no effect on update queries."
)
@click.option(
    "--base64",
    is_flag=True,
    help="Enables base64 encoding of the query parameter for the "
         "SPARQL requests (the response is not touched). "
         "This can be useful in case there is an aggressive firewall between "
         "cmemc and Corporate Memory."
)
@click.option(
    "--parameter", "-p",
    type=(str, str),
    autocompletion=completion.placeholder,
    multiple=True,
    help="In case of a parameterized query (placeholders with the '{{key}}' "
         "syntax), this option fills all placeholder with a given value "
         "before the query is executed."
         "Pairs of placeholder/value need to be given as a tuple 'KEY VALUE'. "
         "A key can be used only once."
)
@click.option(
    "--limit",
    type=int,
    help="Override or set the LIMIT in the executed SELECT query."
)
@click.option(
    "--offset",
    type=int,
    help="Override or set the OFFSET in the executed SELECT query."
)
@click.option(
    "--distinct",
    is_flag=True,
    help="Override the SELECT query by make the result set DISTINCT."
)
@click.option(
    "--timeout",
    type=int,
    help="Set max execution time for query evaluation (in milliseconds)."
)
@click.pass_obj
def execute_command(
        app, queries, accept, no_imports, base64, parameter,
        limit, offset, distinct, timeout
):
    """Execute queries which are loaded from files or the query catalog.

    Queries are identified either by a file path, an URI from the query
    catalog, or a shortened URI (qname, using a default namespace).

    If multiple queries are executed one after the other, the first failing
    query stops the whole execution chain.

    Limitations: All optional parameters (e.g. accept, base64, ...) are
    provided for ALL queries in an execution chain. If you need different
    parameters for each query in a chain, run cmemc multiple times and use
    the logical operators && and || of your shell instead.
    """
    # pylint: disable=too-many-arguments
    placeholder = dict()
    for key, value in parameter:
        if key in placeholder:
            raise ValueError(
                "Parameter can be given only once, "
                "Value for '{}' was given twice.".format(key)
            )
        placeholder[key] = value
    app.echo_debug("Parameter: " + str(placeholder))
    for file_or_uri in queries:
        executed_query = QUERY_CATALOG.get_query(file_or_uri)
        if executed_query is None:
            raise ValueError(
                "{} is neither a (readable) file nor a query URI."
                .format(file_or_uri)
            )
        app.echo_debug("Execute ({}): {} < {}"
                       .format(executed_query.query_type,
                               executed_query.label,
                               executed_query.url)
                       )
        if accept == "default":
            submitted_accept = executed_query.get_default_accept_header()
            app.echo_debug(
                "Accept header set to default value: '{}'"
                .format(submitted_accept)
            )
        else:
            submitted_accept = accept

        results = executed_query.get_results(
            accept=submitted_accept,
            owl_imports_resolution=not no_imports,
            base64_encoded=base64,
            placeholder=placeholder,
            distinct=distinct,
            limit=limit,
            offset=offset,
            timeout=timeout
        )
        app.echo_result(results)


@click.command(name="open")
@click.argument(
    "QUERIES",
    nargs=-1,
    required=True,
    autocompletion=completion.remote_queries_and_sparql_files
)
@click.pass_obj
def open_command(app, queries):
    """Open queries in the editor of the query catalog in your browser.

    With this command, you can open (remote) queries from the query catalog in
    the query editor in your browser (e.g. in order to change them).
    You can also load local query files into the query editor, in order to
    import them into the query catalog.

    The command accepts multiple query URIs or files which results in
    opening multiple browser tabs.
    """
    for file_or_uri in queries:
        opened_query = QUERY_CATALOG.get_query(file_or_uri)
        if opened_query is None:
            raise ValueError(
                "{} is neither a (readable) file nor a query URI."
                .format(file_or_uri)
            )
        open_query_uri = opened_query.get_editor_url()
        app.echo_debug("Open {}: {}".format(file_or_uri, open_query_uri))
        click.launch(open_query_uri)


@click.group()
def query():
    """List, execute or open local and remote SPARQL queries.

    Queries are identified either by a file path, an URI from the query
    catalog or a shortened URI (qname, using a default namespace).

    In order to get a list of queries from the query catalog, use the list
    command. One or more queries can be executed one after the other with the
    execute command. With open command you can jump to the query editor in your
    browser.

    Queries can use a mustache like syntax to specify placeholder for
    parameter values (e.g. {{resourceUri}}). These parameter values need to
    be given as well, before the query can be executed (use the -p option).
    """


query.add_command(execute_command)
query.add_command(list_command)
query.add_command(open_command)



"""Utility functions for CLI interface."""

import os

from cmem.cmempy.dp import proxy  # noqa: I100


def read_rdf_graph_files(directory_path):
    """Read all files from directory_path and output as tuples.

    The tuple format is (filepath, graph_name),
    for example ("/tmp/rdf.nt", "http://example.com")
    """
    rdf_graphs = []
    for root, _, files in os.walk(directory_path):
        for _file in files:
            full_file_path = os.path.join(root, _file)
            graph_file_name = _file + ".graph"
            full_graph_file_name_path = os.path.join(
                root,
                graph_file_name
            )
            if(not _file.endswith(".graph")
               and os.path.exists(full_graph_file_name_path)):
                graph_name = read_file_to_string(
                    full_graph_file_name_path
                ).strip()
                rdf_graphs.append(
                    (full_file_path, graph_name)
                )
    return rdf_graphs


def read_ttl_files(directory_path):
    """Read all files from directory_path.

    Filter the files which ends with .ttl
    Returns list of tuples (full_path_to_file, filename)
    """
    ttl_files = []
    for root, _, files in os.walk(directory_path):
        for _file in files:
            full_file_path = os.path.join(root, _file)
            if _file.endswith(".ttl"):
                ttl_files.append(
                    (full_file_path, _file)
                )
    return ttl_files


def read_file_to_string(file_path):
    """Read file to string."""
    with open(file_path, "rb") as _file:
        return _file.read().decode("utf-8")


def get_graphs(writeable=True, readonly=True):
    """Retrieve list of accessible graphs from DP endpoint.

    readonly=True|writeable=True outputs all graphs
    readonly=False|writeable=True outputs only writeable graphs
    readonly=True|writeable=False outputs graphs without write access
    (but read access)
    """
    all_graphs = proxy.get()['graphs']
    filtered_graphs = []
    for graph in all_graphs:
        if graph['writeable'] and writeable:
            filtered_graphs.append(graph)
        if not graph['writeable'] and readonly:
            filtered_graphs.append(graph)
    return filtered_graphs



"""The main command line interface."""

import json
import os
import sys
import traceback

import click

import requests.exceptions

from cmem.cmemc.cli import completion
from cmem.cmemc.cli.commands import (
    config,
    dataset,
    graph,
    project,
    query,
    vocabulary,
    workflow,
    workspace
)
from cmem.cmemc.cli.context import CONTEXT
from cmem.cmemc.cli.exceptions import (
    InvalidConfiguration
)

try:
    from cmem.cmemc.cli.version import VERSION
except ImportError:
    VERSION = "SNAPSHOT"


PYTHON_VERSION = "{}.{}.{}".format(
    sys.version_info.major,
    sys.version_info.minor,
    sys.version_info.micro
)
PYTHON_EXPECTED = "3.7"
PYTHON_GOT = "{}.{}".format(
    sys.version_info.major,
    sys.version_info.minor
)
if PYTHON_EXPECTED != PYTHON_GOT:
    # test for environment which indicates that we are in completion mode
    if os.getenv("COMP_WORDS", default=None):
        pass
    elif os.getenv("_CMEMC_COMPLETE", default=None):
        pass
    else:
        CONTEXT.echo_warning(
            "Warning: Your are running cmemc under a non-tested python "
            "environment (expected {}, got {})"
            .format(PYTHON_EXPECTED, PYTHON_GOT)
        )


# https://github.com/pallets/click/blob/master/examples/complex/complex/cli.py
CONTEXT_SETTINGS = dict(
    auto_envvar_prefix='CMEMC',
    help_option_names=['-h', '--help']
)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option('-d', '--debug',
              is_flag=True,
              help='Output debug messages and stack traces after errors.')
@click.option('-q', '--quiet',
              is_flag=True,
              help='Suppress any non-error info messages.')
@click.option('--config-file',
              autocompletion=completion.ini_files,
              type=click.Path(
                  readable=True,
                  allow_dash=False,
                  dir_okay=False
              ),
              default=CONTEXT.config_file_default,
              show_default=True,
              help='Use this config file instead of the default one.')
@click.option('-c', '--connection',
              type=click.STRING,
              autocompletion=completion.connections,
              help='Use a specific connection from the config file.')
@click.version_option(
    version=VERSION,
    message="%(prog)s, version %(version)s, "
            "running under python {}".format(PYTHON_VERSION)
)
@click.pass_context
def cli(ctx, debug, quiet, config_file, connection):  # noqa: D403
    """eccenca Corporate Memory Control (cmemc).

    cmemc is the eccenca Corporate Memory Command Line Interface (CLI).

    Available commands are grouped by affecting resource type (such as graph,
    project and query). Each command / group has a separate --help
    screen to get command / group specific documentation.

    In order to see possible commands in a command group, simply
    execute the group command without further parameter (e.g. cmemc project).
    Some commands such as list, import and export can be executed in most
    command groups.
    \b
    Please have a look at the cmemc manual page for more information:

                        https://eccenca.com/go/cmemc
    """
    ctx.obj = CONTEXT
    ctx.obj.set_quiet(quiet)
    ctx.obj.set_debug(debug)
    ctx.obj.set_config_file(config_file)
    ctx.obj.set_connection(connection)


cli.add_command(config.config)
cli.add_command(workspace.workspace)
cli.add_command(graph.graph)
cli.add_command(query.query)
cli.add_command(project.project)
cli.add_command(vocabulary.vocabulary)
cli.add_command(workflow.workflow)
cli.add_command(dataset.dataset)


def main():
    """Start the command line interface."""
    try:
        cli()  # pylint: disable=no-value-for-parameter
    except (
            requests.exceptions.HTTPError,
            requests.exceptions.ConnectionError,
            ValueError,
            IOError,
            InvalidConfiguration
    ) as error:
        CONTEXT.echo_error(str(error))
        CONTEXT.echo_debug(traceback.format_exc())
        # exceptions with response is HTTPError
        try:
            # try to load Problem Details for HTTP API JSON
            details = json.loads(error.response.text)
            error_message = ""
            if 'title' in details:
                error_message += details["title"] + ": "
            if 'detail' in details:
                error_message += details["detail"]
            CONTEXT.echo_error(error_message)
        except (AttributeError, ValueError):
            # is not json or any other issue, output plain response text
            pass
        sys.exit(1)



"""The main command line interface."""
import configparser
from datetime import (
    date,
    datetime
)
import json
from os import environ as env
from os import getenv
from os import makedirs, path

import click

import urllib3

import cmem.cmempy.config as cmempy_config
from cmem.cmemc.cli.exceptions import (InvalidConfiguration)


KNOWN_CONFIG_KEYS = (
    'CMEM_BASE_PROTOCOL',
    'CMEM_BASE_DOMAIN',
    'CMEM_BASE_URI',
    'SSL_VERIFY',
    'REQUESTS_CA_BUNDLE',
    'DP_API_ENDPOINT',
    'DI_API_ENDPOINT',
    'OAUTH_TOKEN_URI',
    'OAUTH_GRANT_TYPE',
    'OAUTH_USER',
    'OAUTH_PASSWORD',
    'OAUTH_CLIENT_ID',
    'OAUTH_CLIENT_SECRET'
)

KNOWN_SECRET_KEYS = (
    'OAUTH_PASSWORD',
    'OAUTH_CLIENT_SECRET'
)


class ApplicationContext():
    """Context of the command line interface."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self, debug=False, quiet=False, connection=''):
        """Initialize main context."""
        self.app_name = "cmemc"
        self.debug = self.set_debug(debug)
        self.quiet = self.set_quiet(quiet)
        self.config_dir = click.get_app_dir(self.app_name)
        if not path.exists(self.config_dir):
            makedirs(self.config_dir)
        self.config_file = self.set_config_file(
            path.join(
                self.config_dir,
                'config.ini'
            )
        )
        self.config_file_default = path.join(
            self.config_dir,
            'config.ini'
        )
        self.config = None
        self.connection = self.set_connection(connection)

    def get_template_data(self):
        """Get the template data dict with vars from the context."""
        data = {}
        today = str(date.today())
        data.update(date=today)
        if self.connection is not None:
            data.update(connection=self.connection.name)
        else:
            data.update(connection="unnamed")
        return data

    def set_debug(self, debug=False):
        """Set debug state."""
        self.debug = debug
        return self.debug

    def set_quiet(self, quiet=False):
        """Set quiets state."""
        self.quiet = quiet
        return self.quiet

    def set_config_file(self, config_file):
        """Set and return the context config file."""
        self.echo_debug('Set config to ' + config_file)
        self.config_file = config_file
        return self.config_file

    def configure_cmempy(self, config):  # noqa: C901
        """Configure the cmempy API to use a new connection."""
        # With or without connection config, we do not want API stdout prints
        env['CMEMPY_IS_CHATTY'] = "False"
        if config is None:
            return
        for key in KNOWN_CONFIG_KEYS:
            # with a loaded config, we delete all known keys from outside
            if key in env:
                env.pop(key)
            if key in config:
                value = str(config[key.lower()])
                env[key] = value
                if key in KNOWN_SECRET_KEYS:
                    self.echo_debug(key + ' set by config')
                else:
                    self.echo_debug(key + ' set by config to ' + value)
        if str(config.get('OAUTH_PASSWORD_ENTRY')).lower() == "true":
            # ask for password hidden and on stderr (in case of piping)
            env['OAUTH_PASSWORD'] = click.prompt(
                'Please enter password for user {}'
                .format(config.get('OAUTH_USER')),
                err=True,
                type=str,
                hide_input=True
            )

        self.echo_debug(
            "CA bundle loaded from {}"
            .format(cmempy_config.get_requests_ca_bundle())
        )
        # If cert validation is disabled, output a warning
        # Also disable library warnings:
        # https://urllib3.readthedocs.io/en/latest/advanced-usage.html
        if not cmempy_config.get_ssl_verify():
            self.echo_warning(
                "SSL verification is disabled (SSL_VERIFY=False)."
            )
            urllib3.disable_warnings()
        return

    def set_connection_from_args(self, args):
        """Set connection and config by manually checking args (completion)."""
        # look for environment and load config
        self.set_config_file(
            getenv("CMEMC_CONFIG_FILE", self.config_file)
        )
        # look for config file in arguments and load config
        found_config_file = False
        for arg in args:
            if found_config_file is True:
                self.set_config_file(arg)
                break
            if arg == "--config-file":
                found_config_file = True
        self.config = self.get_config()
        # look for connection in environment and set connection
        self.set_connection(
            getenv("CMEMC_CONNECTION", '')
        )
        # look for connection in arguments and set connection
        found_connection = False
        for arg in args:
            if found_connection is True:
                self.set_connection(arg)
                return
            if arg in ("-c", "--connection"):
                found_connection = True
        return

    def set_connection(self, section_string):
        """Set connection config section based on section string."""
        self.config = self.get_config()
        self.connection = None
        if section_string is None or section_string == '':
            self.echo_debug('No config given, use API default connection.')
        elif section_string not in self.config:
            message = 'There is no connection "{}" configured in config "{}".'\
                .format(section_string, self.config_file)
            raise InvalidConfiguration(message)
        else:
            self.echo_debug('Use connection config: ' + section_string)
            self.connection = self.config[section_string]
        self.configure_cmempy(self.connection)
        return self.connection

    def get_config_file(self):
        """Check the connection config file."""
        if not path.exists(self.config_file):
            self.echo_warning('Empty config created: {}'
                              .format(self.config_file))
            open(self.config_file, 'a').close()
        self.echo_debug('Config loaded: ' + self.config_file)
        return self.config_file

    def get_config(self):
        """Parse the app connections config."""
        config = configparser.RawConfigParser()
        config_file = self.get_config_file()
        # https://stackoverflow.com/questions/1648517/
        config.read(config_file, encoding='utf-8')
        return config

    @staticmethod
    def echo_warning(message, nl=True):
        """Output a warning message."""
        # pylint: disable=invalid-name
        click.secho(message,
                    fg='yellow',
                    err=True,
                    nl=nl)

    @staticmethod
    def echo_error(message, nl=True, err=True):
        """Output an error message."""
        # pylint: disable=invalid-name
        click.secho(message,
                    fg='red',
                    err=err,
                    nl=nl)

    def echo_debug(self, message):
        """Output a debug message if --debug is enabled."""
        # pylint: disable=invalid-name
        if self.debug:
            click.secho('[{}] {}'.format(str(datetime.now()), message),
                        err=True,
                        dim=True)

    def echo_info(self, message, nl=True):
        """Output an info message, if not suppressed by --quiet."""
        # pylint: disable=invalid-name
        if self.quiet:
            pass
        else:
            click.echo(message, nl=nl)

    def echo_info_json(self, object_):
        """Output a formatted json as info message."""
        # pylint: disable=invalid-name
        if not self.quiet:
            message = json.dumps(object_, indent=2)
            click.echo(message)

    def echo_success(self, message, nl=True):
        """Output success message, if not suppressed by --quiet."""
        # pylint: disable=invalid-name
        if self.quiet:
            pass
        else:
            click.secho(message,
                        fg='green',
                        nl=nl)

    @staticmethod
    def echo_result(message, nl=True):
        """Output result message, can NOT be suppressed by --quiet."""
        # pylint: disable=invalid-name
        click.echo(message, nl=nl)


CONTEXT = ApplicationContext()



"""workflow commands for cmem command line interface."""
import datetime
import json
import sys
import time

import click

import requests

import timeago

from cmem.cmemc.cli import completion
from cmem.cmempy.workflow import get_workflows
from cmem.cmempy.workflow.workflow import (
    get_workflow_editor_uri,
)
from cmem.cmempy.workspace.activities import (
    ACTIVITY_TYPE_EXECUTE_LOCALWORKFLOW,
    VALID_ACTIVITY_STATUS
)
from cmem.cmempy.workspace.activities.taskactivities import (
    get_activities_status
)
from cmem.cmempy.workspace.activities.taskactivity import (
    get_activity_status,
    start_task_activity
)
from cmem.cmempy.workspace.projects.project import get_projects


def _workflows_get_ids():
    """Get a list of workflow IDs."""
    ids = []
    for project_desc in get_projects():
        for workflow_id in get_workflows(project_desc["name"]):
            ids.append(project_desc["name"] + ":" + workflow_id)
    return ids


def workflow_wait_until_finished(project_id, task_id, polling_interval):
    """Poll workflow status until workflow is finished and return status.

    :type project_id: string
    :type task_id: string
    :type polling_interval: int
    """
    while True:
        status = get_activity_status(project_id, task_id)
        # wait until isRunning is false
        if not status['isRunning']:
            break
        time.sleep(polling_interval)
    return status


def _status_command_pre_cmem_20_10(
        app, raw, _filter, workflow_ids
):
    """Get status of workflows without new taskActivitiesStatus API."""
    if workflow_ids == ():
        workflow_ids += tuple(_workflows_get_ids())

    for workflow_id in workflow_ids:
        project_id, task_id = workflow_id.split(":")
        status = get_activity_status(project_id, task_id)
        # filter for running activities
        if _filter == "running"\
                and 'isRunning' in status and not status['isRunning']:
            continue
        # filter for failed activities
        if _filter == "failed"\
                and "failed" in status and not status['failed']:
            continue
        # output raw json if wanted
        if raw:
            app.echo_info(json.dumps(status, indent=2))
        # output single line status
        else:
            app.echo_info("{} ... ".format(workflow_id), nl=False)
            workflow_echo_status(app, status)


def workflow_echo_status(app, status):
    """Print a colored status based on the returned JSON.

    Status can be Idle, Running, Canceling, Waiting, Finished
    isRunning is true for Running, Canceling, Waiting
    canceled only exists sometimes
    """
    # prepare human friendly relative time
    now = datetime.datetime.now()
    stamp = datetime.datetime.fromtimestamp(status["lastUpdateTime"] / 1000)
    time_ago = timeago.format(stamp, now, 'en')
    # prepare message
    if status["statusName"] == status["message"]:
        message = "{} ({})".format(
            status["statusName"],
            time_ago
        )
    else:
        message = "{} ({}, {})".format(
            status["statusName"],
            status["message"],
            time_ago)

    if status["isRunning"]:
        if status["statusName"] == "Running":
            app.echo_warning(message)
        elif status["statusName"] == "Canceling":
            app.echo_warning(message)
        elif status["statusName"] == "Waiting":
            app.echo_warning(message)
        else:
            raise ValueError(
                "statusName is %s, expecting one of: "
                "Running, Canceling or Waiting." % (status["statusName"])
            )
    else:
        # not running can be Idle or Finished
        if "failed" in status and status["failed"]:
            app.echo_error(message, nl=True, err=False)
        elif "cancelled" in status and status["cancelled"]:
            app.echo_warning(message)
        elif status["statusName"] == "Idle":
            app.echo_info(message)
        else:
            # Finished and without failed or canceled status
            app.echo_success(message)


@click.command(name="execute")
@click.option(
    "-a", "--all", "all_",
    is_flag=True,
    help="Execute all available workflows."
)
@click.option(
    "--wait",
    is_flag=True,
    help="Wait for one workflow to complete."
)
@click.option(
    "--polling-interval",
    type=click.IntRange(min=0, max=60),
    show_default=True,
    default=1,
    help="How many seconds to wait between status polls. Status polls are"
         " cheap, so a higher polling interval is most likely not needed."
)
@click.argument(
    "workflow_ids",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.workflow_ids
)
@click.pass_obj
def execute_command(app, all_, wait, polling_interval, workflow_ids):
    """Execute workflow(s).

    With this command, you can start one or more workflows at the same time or
    in a sequence, depending on the result of the predecessor.

    Executing a workflow can be done in two ways: Without --wait just sends
    the starting signal and does not look for the workflow and its result
    (fire and forget). Starting workflows in this way, starts all given
    workflows at the same time.

    The optional --wait option starts the workflows in the same way, but also
    polls the status of a workflow until it is finished. In case of an error of
    a workflow, the next workflow is not started.
    """
    all_workflow_ids = _workflows_get_ids()
    if workflow_ids == () and not all_:
        raise ValueError("Either specify at least one workflow or use the"
                         + " --all option to execute all workflows.")
    if all_:
        workflow_ids = all_workflow_ids

    for workflow_id in workflow_ids:
        if workflow_id not in all_workflow_ids:
            raise ValueError(
                "Workflow '{}' is not available.".format(workflow_id)
            )
        project_id, task_id = workflow_id.split(":")
        app.echo_info("{} ... ".format(workflow_id), nl=False)

        # before we start, we fetch the status
        status = get_activity_status(project_id, task_id)
        if not wait:
            if status['isRunning']:
                # in case of a running workflow, we only output status
                app.echo_info('Already Running')
            else:
                # in case of simple call, we just start and forget
                start_task_activity(project_id, task_id)
                app.echo_info('Started')
        else:
            # in case of --wait, we poll the status until finished
            if status['isRunning']:
                # in case of a running workflow, we only output status
                app.echo_info('Already Running ... ', nl=False)
            else:
                start_task_activity(project_id, task_id)
                app.echo_info('Started ... ', nl=False)

            status = workflow_wait_until_finished(
                project_id, task_id, polling_interval
            )
            # when we have a Finished status, we output it
            workflow_echo_status(app, status)
            # in case of failure, the following workflows are not executed
            if status['failed']:
                sys.exit(1)


@click.command(name="list")
@click.pass_obj
def list_command(app):
    """List available workflow ids."""
    for workflow_id in _workflows_get_ids():
        app.echo_info(workflow_id)


@click.command(name="status")
@click.option(
    "--project", "project_id",
    type=click.STRING,
    autocompletion=completion.project_ids,
    help="The project, from which you want to list the workflows. "
         "Project IDs can be listed with the 'project list' command."
)
@click.option(
    "--raw",
    is_flag=True,
    help="Output raw JSON info."
)
@click.option(
    "--filter", "_filter",
    type=click.Choice(VALID_ACTIVITY_STATUS, case_sensitive=False),
    help="Show only workflows of a specific status."
)
@click.argument(
    "workflow_ids",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.workflow_ids
)
@click.pass_obj
def status_command(app, project_id, raw, _filter, workflow_ids):
    """Get status information of workflow(s)."""
    try:
        workflow_status = get_activities_status(
            status=_filter,
            project_name=project_id,
            activity=ACTIVITY_TYPE_EXECUTE_LOCALWORKFLOW
        )
        for status in workflow_status:
            workflow_id = status["project"] + ":" + status["task"]
            if len(workflow_ids) == 0 or workflow_id in workflow_ids:
                if raw:
                    app.echo_info_json(status)
                else:
                    app.echo_info("{} ... ".format(workflow_id), nl=False)
                    workflow_echo_status(app, status)
    except requests.exceptions.HTTPError:
        app.echo_warning("CMEM 20.10 taskActivitiesStatus API not available.\n"
                         "Falling back to older and more slow API with less "
                         "filter options.")
        _status_command_pre_cmem_20_10(app, raw, _filter, workflow_ids)


@click.command(name="open")
@click.argument(
    "workflow_id",
    type=click.STRING,
    autocompletion=completion.workflow_ids
)
@click.pass_obj
def open_command(app, workflow_id):
    """Open a workflow in your browser."""
    project_id, task_id = workflow_id.split(":")
    workflow_editor_uri = get_workflow_editor_uri()\
        .format(project_id, task_id)
    click.launch(workflow_editor_uri)
    app.echo_debug(workflow_editor_uri)


@click.group()
def workflow():
    """List, execute, open or inspect workflows.

    Workflows are identified by a WORKFLOW_ID. The get a list of existing
    workflows, execute the list command or use tab-completion.
    The WORKFLOW_ID is a concatenation of an PROJECT_ID and a TASK_ID, such as
    "my-project:my-workflow".
    """


workflow.add_command(execute_command)
workflow.add_command(list_command)
workflow.add_command(status_command)
workflow.add_command(open_command)



"""DataIntegration project commands for the cmem command line interface."""

import os
try:
    import pathlib
except ImportError:
    import pathlib2 as pathlib

import click

from jinja2 import Template

from cmem.cmemc.cli import completion
from cmem.cmempy.plugins.marshalling import (
    get_extension_by_plugin,
    get_plugin_by_extension
)
from cmem.cmempy.workspace.projects.export_ import export_project
from cmem.cmempy.workspace.projects.import_ import import_project
from cmem.cmempy.workspace.projects.project import (
    delete_project,
    get_projects,
    make_new_project
)


@click.command(name="list")
@click.pass_obj
def list_command(app):
    """List available projects.

    Outputs a list of project IDs which can be used as reference for
    the project create, delete, export and import commands.
    """
    for project_desc in get_projects():
        app.echo_info(project_desc["name"])


@click.command(name="delete")
@click.option(
    "-a", "--all", "all_",
    is_flag=True,
    help="Delete all projects. "
         "This is a dangerous option, so use it with care.",
)
@click.argument(
    "project_ids",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.project_ids
)
@click.pass_obj
def delete_command(app, all_, project_ids):
    """Delete project(s).

    This deletes existing data integration projects from Corporate Memory.
    Projects will be deleted without prompting!

    Example: cmemc project delete my_project

    Projects can be listed by using the 'cmemc project list' command.
    """
    if project_ids == () and not all_:
        raise ValueError("Either specify at least one project ID "
                         + "or use the --all option to delete all projects.")
    if all_:
        # in case --all is given, a list of project is fetched
        project_ids = (
            [currentProjects["name"] for currentProjects in get_projects()]
        )
    count = len(project_ids)
    current = 1
    for project_id in project_ids:
        app.echo_info("Delete project {}/{}: {} ... "
                      .format(current, count, project_id), nl=False)
        # always fetch the current list of projects
        if project_id in ([_["name"] for _ in get_projects()]):
            delete_project(project_id)
            app.echo_success("done")
        else:
            app.echo_error("not found")
        current = current + 1


@click.command(name="create")
@click.argument(
    "project_ids",
    nargs=-1,
    required=True,
    type=click.STRING
)
@click.pass_obj
def create_command(app, project_ids):
    """Create empty new project(s).

    This creates one or more new projects.
    Existing projects will not be overwritten.

    Example: cmemc project create my_project

    Projects can be listed by using the 'cmemc project list' command.
    """
    count = len(project_ids)
    current = 1
    for project_id in project_ids:
        app.echo_info("Create new project {}/{}: {} ... "
                      .format(current, count, project_id), nl=False)
        app.echo_debug(get_projects())
        # always fetch the current list of projects
        if project_id not in ([_["name"] for _ in get_projects()]):
            make_new_project(project_id)
            app.echo_success("done")
        else:
            app.echo_error("already there")
        current = current + 1


@click.command(name="export")
@click.option(
    "-a", "--all", "all_",
    is_flag=True,
    help="Export all projects.",
)
@click.option(
    "-o", "--overwrite",
    is_flag=True,
    help="Overwrite existing files. This is a dangerous option, "
         "so use it with care.",
)
@click.option(
    "--output-dir",
    default=".",
    show_default=True,
    type=click.Path(
        writable=True,
        file_okay=False
    ),
    # autocompletion=directories,
    help="Directory, where the project files will be created. "
         "This directory will be created as well it does not exists."
)
@click.option(
    "--type", "marshalling_plugin",
    default="xmlZip",
    show_default=True,
    type=click.STRING,
    autocompletion=completion.marshalling_plugins,
    help="Type of the exported project file(s)."
)
@click.option(
    "--filename-template", "-t", "template",
    default="{{date}}-{{connection}}-{{id}}.project",
    show_default=True,
    type=click.STRING,
    autocompletion=completion.project_export_templates,
    help="Template for the export file name(s). "
         "Possible placeholders are (Jinja2): "
         "{{id}} (the project ID), "
         "{{connection}} (from the --connection option) and "
         "{{date}} (the current date as YYYY-MM-DD). "
         "The file suffix will be appended. "
         "Needed directories will be created."
)
@click.argument(
    "project_ids",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.project_ids
)
@click.pass_obj
def export_command(
        app, all_, project_ids, overwrite, marshalling_plugin,
        template, output_dir):
    """Export project(s) to file(s).

    Projects can be exported with different export formats.
    The default type is a zip archive which includes meta data as well
    as dataset resources.
    If more than one project is exported, a file is created for each project.
    By default, these files are created in the current directory and with
    a descriptive name (see --template option default).

    Example: cmemc project export my_project

    Available projects can be listed by using the 'cmemc project list' command.

    You can use the template string to create subdirectories as well:
    cmemc config list | parallel -I% cmemc -c % project export --all
    -t "dump/{{connection}}/{{date}}-{{id}}.project"
    """
    # pylint: disable=too-many-arguments
    if project_ids == () and not all_:
        raise ValueError("Either specify at least one project ID "
                         + "or use the --all option to export all projects.")
    if all_:
        # in case --all is given, a list of project is fetched
        project_ids = ([_["name"] for _ in get_projects()])

    template_data = app.get_template_data()

    count = len(project_ids)
    current = 1
    for project_id in project_ids:
        # prepare the template data and create the actual file incl. suffix
        template_data.update(id=project_id)
        file_name = Template(template).render(template_data)\
            + "." + get_extension_by_plugin(marshalling_plugin)
        # join with given output directory and normalize full path
        export_path = os.path.normpath(os.path.join(output_dir, file_name))
        app.echo_info("Export project {}/{}: {} to {} ... "
                      .format(current, count, project_id, export_path),
                      nl=False)
        if os.path.exists(export_path) and overwrite is not True:
            app.echo_error("file exists")
        else:
            # always fetch the current list of projects
            if project_id in ([_["name"] for _ in get_projects()]):
                export_data = export_project(project_id, marshalling_plugin)
                # output directory is created lazy
                pathlib.Path(
                    os.path.dirname(export_path)
                ).mkdir(parents=True, exist_ok=True)
                # do the export
                with open(export_path, "wb") as export_file:
                    export_file.write(export_data)
                app.echo_success("done")
            else:
                app.echo_error("project not found")
        current = current + 1


@click.command(name="import")
@click.argument(
    "file",
    autocompletion=completion.project_files,
    type=click.Path(
        allow_dash=False,
        dir_okay=False,
        readable=True
    )
)
@click.argument(
    "project_id",
    type=click.STRING,
    autocompletion=completion.project_ids
)
@click.pass_obj
def import_command(app, file, project_id):
    """Import a project from a file.

    Example: cmemc project import my_project.zip my_project
    """
    file_extension = os.path.splitext(file)[1][1:]
    plugin_id = get_plugin_by_extension(file_extension)
    app.echo_info("Import file {} to project {} ... "
                  .format(file, project_id),
                  nl=False)
    if project_id not in ([_["name"] for _ in get_projects()]):
        import_project(project_id, file, plugin_id)
        app.echo_success("done")
    else:
        app.echo_error("already there")


@click.group()
def project():
    """List, import, export, create or delete projects.

    Projects are identified by an PROJECTID. The get a list of existing
    projects, execute the list command or use tab-completion.
    """


project.add_command(list_command)
project.add_command(export_command)
project.add_command(import_command)
project.add_command(delete_command)
project.add_command(create_command)



"""workspace commands for cmem command line interface."""

import os
try:
    import pathlib
except ImportError:
    import pathlib2 as pathlib

import click

from jinja2 import Template

from cmem.cmemc.cli import completion
from cmem.cmempy.workspace import reload_workspace
from cmem.cmempy.workspace.export_ import export
from cmem.cmempy.workspace.import_ import import_workspace


@click.command(name="export")
@click.option(
    "-o", "--overwrite",
    is_flag=True,
    help="Overwrite existing files. "
         "This is a dangerous option, so use it with care.",
)
@click.option(
    "--type", "marshalling_plugin",
    default="xmlZip",
    show_default=True,
    type=click.STRING,
    autocompletion=completion.marshalling_plugins,
    help="Type of the exported workspace file."
)
@click.option(
    "--filename-template", "-t", "template",
    default="{{date}}-{{connection}}.workspace",
    show_default=True,
    type=click.STRING,
    autocompletion=completion.workspace_export_templates,
    help="Template for the export file name. "
         "Possible placeholders are (Jinja2): "
         "{{connection}} (from the --connection option) and "
         "{{date}} (the current date as YYYY-MM-DD). "
         "The file suffix will be appended. "
         "Needed directories will be created."
)
@click.argument(
    "file",
    autocompletion=completion.workspace_files,
    required=False,
    type=click.Path(
        writable=True,
        allow_dash=False,
        dir_okay=False
    )
)
@click.pass_obj
def export_command(app, overwrite, marshalling_plugin, template, file):
    """Export the complete workspace (all projects) to a ZIP file.

    Depending on the requested type, this ZIP contains either a turtle file
    for each project (type rdfTurtle) or a substructure of resource files and
    XML descriptions (type xmlZip).

    The file name is optional and will be generated with by
    the template if absent.
    """
    if file is None:
        # prepare the template data and create the actual file incl. suffix
        template_data = app.get_template_data()
        file = Template(template).render(template_data) + ".zip"
    file = os.path.normpath(file)
    app.echo_info("Export workspace to {} ... ".format(file), nl=False)
    if os.path.exists(file) and overwrite is not True:
        app.echo_error("file exists")
    else:
        # output directory is created lazy
        pathlib.Path(
            os.path.dirname(file)
        ).mkdir(parents=True, exist_ok=True)
        # do the export
        export_data = export(marshalling_plugin)
        with open(file, "wb") as export_file:
            export_file.write(export_data)
        app.echo_success("done")


@click.command(name="import")
@click.option(
    "--type", "marshalling_plugin",
    default="xmlZip",
    show_default=True,
    type=click.STRING,
    autocompletion=completion.marshalling_plugins,
    help="Type of the exported workspace file."
)
@click.argument(
    "file",
    autocompletion=completion.workspace_files,
    type=click.Path(
        readable=True,
        allow_dash=False,
        dir_okay=False
    )
)
@click.pass_obj
def import_command(app, file, marshalling_plugin):
    """Import the workspace from a file."""
    app.echo_info("Import workspace from {} ... "
                  .format(file),
                  nl=False)
    import_workspace(file, marshalling_plugin)
    app.echo_success("done")


@click.command(name="reload")
@click.pass_obj
def reload_command(app):
    """Reload the workspace from the backend."""
    app.echo_info("Reload workspace  ... ", nl=False)
    reload_workspace()
    app.echo_success("done")


@click.group()
def workspace():
    """Import or export the workspace."""


workspace.add_command(export_command)
workspace.add_command(import_command)
workspace.add_command(reload_command)



"""Utility functions for CLI auto-completion functionality."""
# pylint: disable=unused-argument, broad-except
import os

from cmem.cmemc.cli.context import CONTEXT
from cmem.cmemc.cli.utils import get_graphs
from cmem.cmempy.plugins.marshalling import get_marshalling_plugins
from cmem.cmempy.queries import QUERY_CATALOG
from cmem.cmempy.vocabularies import get_vocabularies
from cmem.cmempy.workflow import get_workflows
from cmem.cmempy.workspace import (
    get_task_plugin_description,
    get_task_plugins,
)
from cmem.cmempy.workspace.projects.project import get_projects
from cmem.cmempy.workspace.search import list_items


def add_metadata_parameter(list_=None):
    """Extend a list with metadata keys and key descriptions."""
    if list_ is None:
        list_ = []
    list_.insert(
        0,
        ("description", "Metadata: A description.")
    )
    list_.insert(
        0,
        ("label", "Metadata: A name.")
    )
    return list_


def dataset_parameter(ctx, args, incomplete):
    """Prepare a list of dataset parameter for a dataset type."""
    CONTEXT.set_connection_from_args(args)
    incomplete = incomplete.lower()
    # look if cursor is in value position of the -p option and
    # return nothing in case it is (values are not completed atm)
    if args[len(args) - 2] in ("-p", "--parameter"):
        return []
    # try to determine the dataset type
    dataset_type = None
    for num, arg in enumerate(args):
        if arg == "--type":
            dataset_type = args[num + 1]
    # without type, we know nothing
    if dataset_type is None:
        return []
    plugin = get_task_plugin_description(dataset_type)
    properties = plugin["properties"]
    options = []
    for key in properties:
        description = "{}: {}".format(
            properties[key]["title"],
            properties[key]["description"],
        )
        options.append((key, description))
    options = add_metadata_parameter(options)
    # restrict to search
    options = [
        key for key in options
        if (key[0].lower().find(incomplete.lower()) != -1
            or key[1].lower().find(incomplete.lower()) != -1
            )
    ]
    return options


def dataset_types(ctx, args, incomplete):
    """Prepare a list of dataset types."""
    CONTEXT.set_connection_from_args(args)
    incomplete = incomplete.lower()
    options = []
    plugins = get_task_plugins()
    for plugin_id in plugins:
        plugin = plugins[plugin_id]
        title = plugin["title"]
        description = "{}: {}".format(
            title,
            plugin["description"].partition("\n")[0]
        )
        if plugin["taskType"] == "Dataset" and (
                title.lower().find(incomplete.lower()) != -1
                or description.lower().find(incomplete.lower()) != -1):
            options.append(
                (
                    plugin_id,
                    description
                )
            )
    return options


def dataset_ids(ctx, args, incomplete):
    """Prepare a list of projectid:datasetid dataset identifier."""
    CONTEXT.set_connection_from_args(args)
    options = []
    results = list_items(item_type="dataset")
    datasets = results["results"]
    for _ in datasets:
        options.append(
            (
                _["projectId"] + r"\:" + _["id"],
                _["label"]
            )
        )
    options = [
        dataset for dataset in options
        if dataset not in args
        # to-complete string can match ID or label
        and (dataset[0].lower().find(
            incomplete.lower()
        ) != -1 or dataset[1].lower().find(
            incomplete.lower()) != -1
        )
    ]
    return options


def vocabularies(ctx, args, incomplete, filter_="all"):
    """Prepare a list of vocabulary graphs for auto-completion."""
    CONTEXT.set_connection_from_args(args)
    incomplete = incomplete.lower()
    options = []
    try:
        vocabs = get_vocabularies(filter_=filter_)
    except Exception:
        # if something went wrong, die silently
        return []
    for _ in vocabs:
        url = _["iri"].replace(":", r"\:")
        label = _["vocabularyLabel"] or "Vocabulary in graph " + url
        if url in args:
            continue
        if incomplete != "":
            if url.lower().find(incomplete) != -1 \
                    or label.lower().find(incomplete) != -1:
                options.append((url, label))
        else:
            options.append((url, label))
    return options


def installed_vocabularies(ctx, args, incomplete):
    """Prepare a list of installed vocabulary graphs."""
    return vocabularies(ctx, args, incomplete, filter_="installed")


def installable_vocabularies(ctx, args, incomplete):
    """Prepare a list of installable vocabulary graphs."""
    return vocabularies(ctx, args, incomplete, filter_="installable")


def file_list(incomplete="", suffix="", description=""):
    """Prepare a list of files with specific parameter."""
    if os.path.isdir(incomplete):
        # given string is a directory, so we scan this directory and
        # add it as a prefix
        directory = incomplete
        incomplete = ""
        prefix = os.path.realpath(incomplete) + "/"
    else:
        # given string is NOT a directory so we just scan the current
        # directory
        directory = os.getcwd()
        prefix = ""
    options = []
    for file_name in os.listdir(directory):
        if os.path.isfile(file_name) \
                and file_name.endswith(suffix) \
                and file_name.lower().find(incomplete.lower()) != -1:
            options.append((prefix + file_name, description))
    return sorted(options, key=lambda x: x[1].casefold())


def dataset_files(ctx, args, incomplete):
    """Prepare a list of SPARQL files."""
    return file_list(
        incomplete=incomplete,
        suffix=".csv",
        description="CSV Dataset resource"
    ) + file_list(
        incomplete=incomplete,
        suffix=".xlsx",
        description="Excel Dataset resource"
    ) + file_list(
        incomplete=incomplete,
        suffix=".xml",
        description="XML Dataset resource"
    ) + file_list(
        incomplete=incomplete,
        suffix=".json",
        description="JSON Dataset resource"
    ) + file_list(
        incomplete=incomplete,
        suffix=".ttl",
        description="RDF file Dataset resource"
    ) + file_list(
        incomplete=incomplete,
        suffix=".zip",
        description="multiCsv Dataset resource"
    ) + file_list(
        incomplete=incomplete,
        suffix=".orc",
        description="Apache ORC Dataset resource"
    )


def project_files(ctx, args, incomplete):
    """Prepare a list of workspace files."""
    return file_list(
        incomplete=incomplete,
        suffix=".project.zip",
        description="eccenca Corporate Memory project backup file"
    )


def ini_files(ctx, args, incomplete):
    """Prepare a list of workspace files."""
    return file_list(
        incomplete=incomplete,
        suffix=".ini",
        description="INI file"
    )


def workspace_files(ctx, args, incomplete):
    """Prepare a list of workspace files."""
    return file_list(
        incomplete=incomplete,
        suffix=".workspace.zip",
        description="eccenca Corporate Memory workspace backup file"
    )


def sparql_files(ctx, args, incomplete):
    """Prepare a list of SPARQL files."""
    return file_list(
        incomplete=incomplete,
        suffix=".sparql",
        description="SPARQL query file"
    ) + file_list(
        incomplete=incomplete,
        suffix=".rq",
        description="SPARQL query file"
    )


def triple_files(ctx, args, incomplete):
    """Prepare a list of triple files."""
    return file_list(
        incomplete=incomplete,
        suffix=".ttl",
        description="RDF Turtle file"
    ) + file_list(
        incomplete=incomplete,
        suffix=".nt",
        description="RDF NTriples file"
    )


def placeholder(ctx, args, incomplete):
    """Prepare a list of placeholder from the to-be executed queries."""
    # look if cursor is in value position of the -p option and
    # return nothing in case it is (values are not completed atm)
    if args[len(args) - 2] in ("-p", "--parameter"):
        return []
    # setup configuration
    CONTEXT.set_connection_from_args(args)
    # extract placeholder from given queries in the command line
    options = []
    for num, arg in enumerate(args):
        query = QUERY_CATALOG.get_query(arg)
        if query is not None:
            options.extend(
                list(query.get_placeholder_keys())
            )
    # look for already given parameter in the arguments and remove them from
    # the available options
    for num, arg in enumerate(args):
        if num - 1 > 0 and args[num - 1] in ("-p", "--parameter"):
            options.remove(arg)
    # remove all non-matching placeholder if incomplete is given
    options = [
        p for p in options
        if p.lower().find(incomplete.lower()) != -1
    ]
    # additional debug output options (needed for all completion functions)
    # options.append("incomplete: -> " + incomplete + " <-")
    # options.append("args: -> " + str(args) + " <-")
    # options.append("len(args): -> " + str(len(args)) + " <-")
    return sorted(set(options), key=str.casefold)


def remote_queries(ctx, args, incomplete):
    """Prepare a list of query URIs."""
    CONTEXT.set_connection_from_args(args)
    incomplete = incomplete.lower()
    options = []
    for _, query in QUERY_CATALOG.get_queries().items():
        url = query.short_url.replace(":", r"\:")
        label = query.label
        if incomplete != "":
            if url.lower().find(incomplete) != -1 \
                    or label.lower().find(incomplete) != -1:
                options.append((url, label))
        else:
            options.append((url, label))
    return sorted(options, key=lambda x: x[1].casefold())


def remote_queries_and_sparql_files(ctx, args, incomplete):
    """Prepare a list of named queries, query files and directories."""
    remote = remote_queries(ctx, args, incomplete)
    files = sparql_files(ctx, args, incomplete)
    return remote + files


def workflow_ids(ctx, args, incomplete):
    """Prepare a list of projectid:taskid workflow identifier."""
    CONTEXT.set_connection_from_args(args)
    options = []
    for project_desc in get_projects():
        for workflow_id in get_workflows(project_desc["name"]):
            options.append(project_desc["name"] + ":" + workflow_id)
    options = [
        workflow for workflow in options
        if workflow not in args
        and workflow.lower().find(incomplete.lower()) != -1
    ]
    return sorted(options, key=str.casefold)


def marshalling_plugins(ctx, args, incomplete):
    """Prepare a list of supported workspace/project import/export plugins."""
    CONTEXT.set_connection_from_args(args)
    options = get_marshalling_plugins()
    if "description" in options[0].keys():
        return [(_["id"], _["description"]) for _ in options]
    # in case, no descriptions are available, labels are fine as well
    return [(_["id"], _["label"]) for _ in options]


def project_ids(ctx, args, incomplete):
    """Prepare a list of project IDs for auto-completion."""
    CONTEXT.set_connection_from_args(args)
    try:
        catalog = get_projects()
    except Exception:
        # if something went wrong, die silently
        return []
    # strip down to an ID only list and filter for already given IDs in args
    # remove IDs where typed string is found in (not begins with)
    options = [
        project['name'] for project
        in catalog
        if project['name'] not in args
        and project['name'].lower().find(incomplete.lower()) != -1
    ]
    return sorted(options, key=str.casefold)


def graph_uris(ctx, args, incomplete):
    """Prepare a list of graphs for auto-completion."""
    CONTEXT.set_connection_from_args(args)
    try:
        graphs = get_graphs()
    except Exception:
        # if something went wrong, die silently
        return []
    # strip down to an IRI only list and filter for already given IRIs in args
    # remove IRIs where typed string is found in (not begins with)
    options = [
        graph['iri'] for graph
        in graphs
        if graph['iri'] not in args
        and graph['iri'].lower().find(incomplete.lower()) != -1
    ]
    return sorted(options, key=str.casefold)


def writable_graph_uris(ctx, args, incomplete):
    """Prepare a list of writable graphs for auto-completion."""
    # seeAlso: https://github.com/pallets/click/issues/942
    CONTEXT.set_connection_from_args(args)
    try:
        graphs = get_graphs(writeable=True, readonly=False)
    except Exception:
        # if something went wrong, die silently
        return []
    # strip down to an IRI only list and filter for already given IRIs in args
    graphs = [graph['iri'] for graph in graphs if graph['iri'] not in args]
    # remove IRIs where typed string is found in (not begins with)
    graphs = [
        iri for iri in graphs
        if iri.lower().find(incomplete.lower()) != -1
    ]
    return sorted(graphs, key=str.casefold)


def connections(ctx, args, incomplete):
    """Prepare a list of config connections for auto-completion."""
    # since ctx does not have an obj here, we re-create the object
    CONTEXT.set_connection_from_args(args)
    options = []
    for section in CONTEXT.config.sections():
        if section.lower().find(incomplete.lower()) != -1:
            options.append(section)
    return sorted(options, key=str.casefold)


def project_export_templates(ctx, args, incomplete):
    """Prepare a list of example templates for the project export command."""
    examples = [
        (
            "{{id}}",
            "Example: a plain file name"
        ),
        (
            "{{date}}-{{connection}}-{{id}}.project",
            "Example: a more descriptive file name"),
        (
            "dumps/{{connection}}/{{id}}/{{date}}.project",
            "Example: a whole directory tree"
        )
    ]
    return examples


def workspace_export_templates(ctx, args, incomplete):
    """Prepare a list of example templates for the workspace export command."""
    examples = [
        (
            "workspace",
            "Example: a plain file name"
        ),
        (
            "{{date}}-{{connection}}.workspace",
            "Example: a more descriptive file name"),
        (
            "dumps/{{connection}}/{{date}}.workspace",
            "Example: a whole directory tree"
        )
    ]
    return examples



"""graph commands for cmem command line interface."""

import hashlib
import json
import os
try:
    import pathlib
except ImportError:
    import pathlib2 as pathlib

from six.moves.urllib.parse import quote

import click

from jinja2 import Template

from cmem.cmemc.cli import completion
from cmem.cmemc.cli.utils import (
    get_graphs,
    read_rdf_graph_files
)
from cmem.cmempy.config import get_cmem_base_uri
from cmem.cmempy.dp.authorization import refresh
from cmem.cmempy.dp.proxy import graph as graph_api
from cmem.cmempy.dp.proxy.sparql import get as sparql_api


def count_graph(graph_iri):
    """Count triples in a graph and return integer."""
    query = "SELECT (COUNT(*) as ?triples)" +\
            " FROM <" + graph_iri + ">" +\
            " WHERE { ?s ?p ?o }"
    result = json.loads(sparql_api(query, owl_imports_resolution=False))
    count = result["results"]["bindings"][0]["triples"]["value"]
    return int(count)


def _get_graph_to_file(
        graph_iri, file_path,
        app,
        numbers=None,
        overwrite=True):
    """Request a single graph to a single file.

    numbers is a tupel of current and count (for output only).
    """
    if os.path.exists(file_path):
        if overwrite is True:
            app.echo_warning(
                "Output file {} does exist: will overwrite it."
                .format(file_path)
            )
        else:
            app.echo_warning(
                "Output file {} does exist: will append to it."
                .format(file_path)
            )
    if numbers is not None:
        running_number, count = numbers
        if running_number is not None and count is not None:
            app.echo_info(
                "Export graph {}/{}: {} to {} ... "
                .format(running_number, count, graph_iri, file_path),
                nl=False
            )
    # create and write the .ttl content file
    if overwrite is True:
        triple_file = click.open_file(file_path, "wb")
    else:
        triple_file = click.open_file(file_path, "ab")
    # TODO: use a streaming approach here
    data = graph_api.get(graph_iri).content
    triple_file.write(data)
    if numbers is not None:
        app.echo_success("done")


@click.command(name="list")
@click.option(
    "--raw",
    is_flag=True,
    help="Outputs raw JSON."
)
@click.option(
    "--filter", "filter_",
    type=click.Choice(
        ["readonly", "writeable"],
        case_sensitive=True
    ),
    help="Filter list based on access conditions."
)
@click.pass_obj
def list_command(app, raw, filter_):
    """List accessible graphs."""
    if filter_ == "writeable":
        graphs = get_graphs(writeable=True, readonly=False)
    elif filter_ == "readonly":
        graphs = get_graphs(writeable=False, readonly=True)
    else:
        graphs = get_graphs(writeable=True, readonly=True)

    if raw:
        app.echo_result(graphs)
    else:
        for graph_desc in graphs:
            app.echo_result(graph_desc["iri"])


# pylint: disable=too-many-arguments
@click.command(name="export")
@click.option(
    "-a", "--all", "all_",
    is_flag=True,
    help="Export all (readable) graphs"
)
@click.option(
    "--output-dir",
    type=click.Path(
        writable=True,
        file_okay=False
    ),
    help="Export to this directory"
)
@click.option(
    "--output-file",
    type=click.Path(
        writable=True,
        allow_dash=True,
        dir_okay=False
    ),
    default="-",
    show_default=True,
    autocompletion=completion.triple_files,
    help="Export to this file"
)
@click.option(
    "--filename-template", "-t", "template",
    default="{{hash}}",
    show_default=True,
    type=click.STRING,
    help="Template for the export file name(s). "
         "Used together with --output-dir. "
         "Possible placeholders are (Jinja2): "
         "{{hash}} (the sha256 hash of the graph URI)"
         "{{connection}} (from the --connection option) and "
         "{{date}} (the current date as YYYY-MM-DD). "
         "The file suffix will be appended. "
         "Needed directories will be created."
)
@click.argument(
    "iris",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.graph_uris
)
@click.pass_obj
def export_command(app, all_, iris, output_dir, output_file, template):
    """Export graph(s) as NTriples to stdout (-), file or directory.

    In case of file export, data from all selected graphs will be concatenated
    in one file.
    In case of directory export, .graph and .nt files will be created
    for each graph.
    """
    if iris == () and not all_:
        raise ValueError("Either specify at least one graph "
                         + "IRI or use the --all option to export all graphs.")
    if all_:
        # in case --all is given,
        # list of graphs is filled with all available graph IRIs
        iris = ([iri["iri"] for iri in get_graphs()])

    template_data = app.get_template_data()

    count: int = len(iris)
    current: int = 1
    app.echo_debug("graph count is " + str(count))
    if output_dir:
        # output directory set
        app.echo_debug("output is directory")
        if not os.path.exists(output_dir):
            app.echo_warning("Output directory does not exist: "
                             + "will create it.")
            os.makedirs(output_dir)
        # in case a directory is the target,
        # the graph/nt file structure is created
        for iri in iris:
            template_data.update(
                hash=hashlib.sha256(iri.encode("utf-8")).hexdigest()
            )
            triple_file_name = Template(template)\
                .render(template_data) + ".ttl"
            # join with given output directory and normalize full path
            triple_file_name = os.path.normpath(
                os.path.join(
                    output_dir,
                    triple_file_name
                )
            )
            graph_file_name = triple_file_name + ".graph"
            # output directory is created lazy
            pathlib.Path(
                os.path.dirname(triple_file_name)
            ).mkdir(parents=True, exist_ok=True)
            # create and write the .ttl.graph meta data file
            graph_file = click.open_file(graph_file_name, "w")
            graph_file.write(iri + "\n")
            _get_graph_to_file(
                iri, triple_file_name,
                app,
                numbers=(current, count)
            )
            current = current + 1
    else:
        # no output directory set -> file export
        if output_file == "-":
            # in case a file is stdout,
            # all triples from all graphs go in and other output is suppressed
            app.echo_debug("output is stdout")
            for iri in iris:
                _get_graph_to_file(iri, output_file, app)
        else:
            # in case a file is given, all triples from all graphs go in
            app.echo_debug("output is file")
            for iri in iris:
                _get_graph_to_file(
                    iri, output_file,
                    app,
                    numbers=(current, count),
                    overwrite=False
                )
                current = current + 1


@click.command(name="import")  # noqa: C901
@click.option(
    "--replace",
    is_flag=True,
    help="Replace (overwrite) original graph data."
)
@click.argument(
    "input_path",
    required=True,
    autocompletion=completion.triple_files,
    type=click.Path(
        allow_dash=False,
        readable=True
    )
)
@click.argument(
    "iri",
    type=click.STRING,
    required=False,
    autocompletion=completion.graph_uris
)
@click.pass_obj
def import_command(app, input_path, replace, iri):
    """Import graph(s) to the store.

    If input is an directory, it scans for file-pairs such as xxx.ttl and
    xxx.ttl.graph where xxx.ttl is the actual triples file and xxx.ttl.graph
    contains the graph IRI as one string: "https://mygraph.de/xxx/".
    If input is a file, content will be uploaded to IRI.
    If --replace is set, the data will be overwritten,
    if not, it will be added.
    """
    # is an array of tuples like this [('path/to/triple.file', 'graph IRI')]
    graphs: list
    if os.path.isdir(input_path):
        if iri is not None:
            raise ValueError("Either specify an input file AND a graph IRI "
                             + "or an input directory ONLY.")
        # in case a directory is the source,
        # the graph/nt file structure is crawled
        graphs = read_rdf_graph_files(input_path)
    elif os.path.isfile(input_path):
        if iri is None:
            raise ValueError("Either specify an input file AND a graph IRI "
                             + "or an input directory ONLY.")
        graphs = [(input_path, iri)]
    else:
        # TODO: support for stdin stream
        raise NotImplementedError(
            "Input from special files "
            + "(socket, FIFO, device file) is not supported."
        )

    processed_graphs: set = set()
    count: int = len(graphs)
    current: int = 1
    for (triple_file, graph_iri) in graphs:
        app.echo_info("Import file {}/{}: {} from {} ... "
                      .format(current, count, graph_iri, triple_file),
                      nl=False)
        # prevents re-replacing of graphs in a single run
        _replace = False if graph_iri in processed_graphs else replace
        graph_api.post(graph_iri, triple_file, replace=_replace)
        app.echo_success("replaced" if _replace else "added")
        # refresh access conditions in case of dropped AC graph
        if graph_iri == refresh.AUTHORIZATION_GRAPH_URI:
            refresh.get()
            app.echo_debug("Access conditions refreshed.")
        processed_graphs.add(graph_iri)
        current += 1


@click.command(name="delete")
@click.option(
    "-a", "--all", "all_",
    is_flag=True,
    help="Drop all (writeable) graphs"
)
@click.argument(
    "iris",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.writable_graph_uris
)
@click.pass_obj
def delete_command(app, all_, iris):
    """Delete graph(s) from the store."""
    if iris == () and not all_:
        raise ValueError("Either specify at least one graph IRI "
                         + "or use the --all option to drop all graphs.")
    if all_:
        # in case --all is given,
        # list of graphs is filled with all available graph IRIs
        iris = (
            [iri["iri"] for iri in get_graphs(writeable=True, readonly=False)]
        )

    count: int = len(iris)
    current: int = 1
    for iri in iris:
        app.echo_info("Drop graph {}/{}: {} ... "
                      .format(current, count, iri), nl=False)
        graph_api.delete(iri)
        app.echo_success("done")
        # refresh access conditions in case of dropped AC graph
        if iri == refresh.AUTHORIZATION_GRAPH_URI:
            refresh.get()
            app.echo_debug("Access conditions refreshed.")
        current = current + 1


@click.command(name="open")
@click.argument(
    "iri",
    type=click.STRING,
    autocompletion=completion.graph_uris
)
@click.pass_obj
def open_command(app, iri):
    """Open / explore a graph in the browser."""
    explore_uri = get_cmem_base_uri() + "/explore?graph=" + quote(iri)
    click.launch(explore_uri)
    app.echo_debug(explore_uri)


@click.command(name="count")
@click.option(
    "-a", "--all", "all_",
    is_flag=True,
    help="Count all graphs"
)
@click.option(
    "-s", "--summarize",
    is_flag=True,
    help="Display only a sum of all counted graphs together"
)
@click.argument(
    "iris",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.graph_uris
)
@click.pass_obj
def count_command(app, all_, summarize, iris):
    """Count triples in graph(s).

    This command lists graphs with their triple count.
    Counts are done without following imported graphs.
    """
    if iris == () and not all_:
        raise ValueError("Either specify at least one graph IRI "
                         + "or use the --all option to count all graphs.")
    if all_:
        # in case --all is given,
        # list of graphs is filled with all available graph IRIs
        iris = (
            [iri["iri"] for iri in get_graphs()]
        )

    count: int
    overall_sum: int = 0
    current: int = 1
    for iri in iris:
        count = count_graph(iri)
        overall_sum = overall_sum + count
        current = current + 1
        if not summarize:
            app.echo_result("{} {}".format(str(count), iri))
    if summarize:
        app.echo_result(overall_sum)


@click.group()
def graph():
    """List, import, export, delete or open graphs.

    Graphs are identified by an IRI. The get a list of existing graphs,
    execute the list command or use tab-completion.
    """


graph.add_command(count_command)
graph.add_command(list_command)
graph.add_command(export_command)
graph.add_command(delete_command)
graph.add_command(import_command)
graph.add_command(open_command)



"""configuration commands for cmem command line interface."""
import click

from cmem.cmemc.cli import completion
from cmem.cmempy.health import (
    di_is_healthy, dp_is_healthy, get_di_version, get_dp_version
)


@click.command(name="list")
@click.pass_obj
def list_command(app):
    """List configured CMEM connections."""
    for section_string in app.config:
        if section_string != "DEFAULT":
            app.echo_result(section_string)


@click.command(name="edit")
@click.pass_obj
def edit_command(app):
    """Edit the user-scope configuration file."""
    app.echo_info("Open editor for config file " + app.config_file)
    click.edit(filename=app.config_file)


@click.command(name="check")
@click.argument(
    "configs",
    nargs=-1,
    type=click.STRING,
    autocompletion=completion.connections
)
@click.option(
    "-a", "--all", "all_",
    is_flag=True,
    help="Export all (readable) graphs"
)
@click.pass_obj
def check_command(app, configs, all_):
    """Check the status of deployment."""
    if all_:
        # in case --all is given,
        # list of configs is filled with all available configs
        configs = ([_ for _ in app.config if _ != "DEFAULT"])
    for config_string in configs:
        app.echo_debug("Now health check for " + config_string)
        app.set_connection(config_string)
        app.echo_success("[{}] ".format(config_string))
        # check DP
        app.echo_result("DataPlatform ", nl=False)
        if dp_is_healthy():
            app.echo_result(get_dp_version() + " ... ", nl=False)
            app.echo_success("UP")
        else:
            app.echo_result(get_dp_version() + " ... ", nl=False)
            app.echo_error("DOWN")
        # check DI
        app.echo_result("DataIntegration ", nl=False)
        if di_is_healthy():
            app.echo_result(get_di_version() + " ... ", nl=False)
            app.echo_success("UP")
        else:
            app.echo_result(get_di_version() + " ... ", nl=False)
            app.echo_error("DOWN")


@click.group()
def config():
    u"""List, edit and check configurations.

    Configurations are identified by the section identifier in the
    config file. Each configuration represent a Corporate Memory deployment
    with its specific access method as well as credentials.

    A minimal configuration which uses client credentials has the following
    entries:

    \b
    [example.org]
    CMEM_BASE_URI=https://cmem.example.org/
    OAUTH_GRANT_TYPE=client_credentials
    OAUTH_CLIENT_ID=cmem-service-account
    OAUTH_CLIENT_SECRET=my-secret-account-pass

    In addition to that, the following config parameters can be used as well:

    \b
    SSL_VERIFY=False    - for ignoring certificate issues (not recommended)
    DP_API_ENDPOINT=URL - to point to a non-standad DataPlatform location
    DI_API_ENDPOINT=URL - to point to a non-standad DataIntegration location
    OAUTH_TOKEN_URI=URL - to point to an external IdentityProvider location
    OAUTH_USER=username - username (only if OAUTH_GRANT_TYPE=password)
    OAUTH_PASSWORD=pass - password (only if OAUTH_GRANT_TYPE=password)

    In order to request passwords on start, you can use the following
    parameter instead the PASSWORD parameter: OAUTH_PASSWORD_ENTRY=True
    OAUTH_CLIENT_SECRET_ENTRY=True.
    """


config.add_command(list_command)
config.add_command(check_command)
config.add_command(edit_command)
